from __future__ import annotations

from typing import Any, List

import click

from hyperfocus import formatter
from hyperfocus.database.models import Task


def echo(text: str):
    click.secho(text)


def task(task: Task, show_details: bool = False, show_prefix: bool = False):
    formatted_task = formatter.task(
        task=task, show_details=show_details, show_prefix=show_prefix
    )
    echo(text=formatted_task)


def tasks(tasks: List[Task], newline: bool = False):
    formatted_tasks = formatter.tasks(tasks=tasks, newline=newline)
    echo(text=formatted_tasks)


def notification(text: str, event: str, status: formatter.NotificationLevel):
    formatted_notification = formatter.notification(
        text=text, event=event, status=status
    )
    echo(text=formatted_notification)


def info(text: str, event: str):
    notification(text=text, event=event, status=formatter.NotificationLevel.INFO)


def success(text: str, event: str):
    notification(text=text, event=event, status=formatter.NotificationLevel.SUCCESS)


def warning(text: str, event: str):
    notification(text=text, event=event, status=formatter.NotificationLevel.WARNING)


def error(text: str, event: str):
    notification(text=text, event=event, status=formatter.NotificationLevel.ERROR)


def ask(text: str, **kwargs) -> Any:
    formatted_prompt = formatter.prompt(text=text)
    return click.prompt(text=formatted_prompt, **kwargs)


def confirm(text: str, **kwargs) -> Any:
    formatted_prompt = formatter.prompt(text=text)
    return click.confirm(text=formatted_prompt, **kwargs)



from __future__ import annotations

from pathlib import Path

import click

from hyperfocus import __app_name__


CONFIG_DIR = Path(click.get_app_dir(__app_name__))
CONFIG_PATH = CONFIG_DIR / "config.ini"

DEFAULT_DB_PATH = Path.home() / f".{__app_name__}.sqlite"



from __future__ import annotations

import click
from click import Parameter


class NotRequiredIf(click.Argument):
    def __init__(self, *args, not_required_if: list[str], **kwargs):
        self.not_required_if = not_required_if
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt: bool = self.name in opts

        for not_required_opt in self.not_required_if:
            other_opt: bool = not_required_opt in opts
            if all([current_opt, other_opt]):
                self.required = False

        return super().handle_parse_result(ctx, opts, args)


class NotRequired(click.Option):
    def __init__(self, *args, not_required: list[str], **kwargs):
        self.not_required = not_required
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        params = {p.name: p for p in ctx.command.get_params(ctx)}
        current_opt: bool = self.name in opts

        for not_required_opt in self.not_required:
            other_option: bool = not_required_opt in opts and opts[not_required_opt]
            if all([current_opt, other_option]):
                param: Parameter = params[not_required_opt]
                raise click.exceptions.UsageError(
                    f"Unnecessary {param.param_type_name} "
                    f"'{param.human_readable_name}' provided",
                    ctx=ctx,
                )

        return super().handle_parse_result(ctx, opts, args)



from __future__ import annotations

import click

from hyperfocus.config.config import Config
from hyperfocus.hyf_click.error_handler import hyf_error_handler
from hyperfocus.utils import wrap_methods


@wrap_methods(hyf_error_handler, ["make_context", "invoke"])
class HyfGroup(click.Group):
    def get_command(self, ctx: click.Context, cmd_name: str):
        cmd = click.Group.get_command(self, ctx=ctx, cmd_name=cmd_name)
        if cmd is not None:
            return cmd

        config = Config.load()
        option = f"alias.{cmd_name}"
        if option in config:
            cmd_name = config[option]

        return click.Group.get_command(self, ctx, cmd_name)

    def resolve_command(
        self, ctx: click.Context, args: list[str]
    ) -> tuple[str | None, click.Command | None, list[str]]:
        # always return the command's name, not the alias
        _, cmd, args = super().resolve_command(ctx, args)
        cmd_name = cmd if cmd is None else cmd.name
        return cmd_name, cmd, args



from __future__ import annotations

from configparser import ConfigParser
from pathlib import Path

from hyperfocus.config.exceptions import ConfigFileError


class ConfigFile:
    def __init__(self, path: str | Path) -> None:
        if isinstance(path, str):
            path = Path(path)
        self._path = path
        self._parser = ConfigParser()

    @property
    def path(self) -> str:
        return str(self._path)

    def exists(self) -> bool:
        return self._path.exists()

    def read(self) -> dict[str, dict]:
        self._parser.read(self._path)
        return {s: dict(self._parser.items(s)) for s in self._parser.sections()}

    def write(self, config: dict[str, dict]) -> None:
        for section, items in config.items():
            self._parser[section] = items

        try:
            if not self.exists():
                self._path.touch(mode=0o600)

            with self._path.open("w") as file:
                self._parser.write(file)
        except OSError as error:
            raise ConfigFileError(str(error)) from error



from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from typing import List, Optional

from hyperfocus.database.models import DailyTracker, Task, TaskStatus


class DailyTrackerServiceBase(ABC):
    @property
    @abstractmethod
    def date(self) -> datetime.date:
        raise NotImplementedError

    @abstractmethod
    def add_task(self, title: str, details: str | None = None) -> Task:
        raise NotImplementedError

    @abstractmethod
    def get_task(self, task_id: int) -> Task | None:
        raise NotImplementedError

    @abstractmethod
    def get_tasks(self, exclude: list[TaskStatus] | None = None) -> list[Task]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def update_task(task: Task, status: TaskStatus) -> None:
        raise NotImplementedError


class DailyTrackerService(DailyTrackerServiceBase):
    def __init__(self, daily_tracker: DailyTracker, is_a_new_day: bool = False) -> None:
        self._base = daily_tracker
        self.is_a_new_day = is_a_new_day

    @classmethod
    def from_date(cls, date: datetime.date) -> DailyTrackerService:
        daily_tracker, is_a_new_day = DailyTracker.get_or_create(date=date)
        return cls(daily_tracker=daily_tracker, is_a_new_day=is_a_new_day)

    @property
    def date(self) -> datetime.date:
        return self._base.date

    def add_task(self, title: str, details: Optional[str] = None) -> Task:
        details = details.strip() if isinstance(details, str) else details
        task = Task(
            id=self._base.next_task_id,
            title=title.strip(),
            details=details,
            status=TaskStatus.TODO,
            daily_tracker=self._base,
        )
        task.save()

        self._base.task_increment += 1
        self._base.save()

        return task

    def get_task(self, task_id: int) -> Task | None:
        return Task.get_or_none(Task.id == task_id, Task.daily_tracker == self._base)

    def get_tasks(self, exclude: list[TaskStatus] | None = None) -> List[Task]:
        exclude = exclude or []
        tasks = [task for task in self._base.tasks if task.status not in exclude]
        return tasks

    @staticmethod
    def update_task(task: Task, status: TaskStatus) -> None:
        task.status = status
        task.save()


class NullDailyTrackerService(DailyTrackerServiceBase):
    null_date = datetime.datetime(1970, 1, 1).date()

    def __init__(self, date: datetime.date | None = None) -> None:
        self._date = date or self.null_date
        self.daily_tracker = None
        self.new_day = None

    @property
    def date(self) -> datetime.date:
        return self._date

    def add_task(self, title: str, details: str | None = None) -> Task:
        pass

    def get_task(self, task_id: int) -> Task | None:
        pass

    def get_tasks(self, exclude: list[TaskStatus] | None = None) -> list[Task]:
        return []

    @staticmethod
    def update_task(task: Task, status: TaskStatus) -> None:
        pass


class PastTrackerService:
    def __init__(self, current_day: DailyTrackerService) -> None:
        self._current_day = current_day

    def get_previous_day(self) -> DailyTrackerServiceBase:
        date = self._current_day.date
        previous_daily_tracker = (
            DailyTracker.select()
            .where(DailyTracker.date < date)
            .order_by(DailyTracker.date.desc())
            .get_or_none()
        )
        if not previous_daily_tracker:
            return NullDailyTrackerService()

        return DailyTrackerService(daily_tracker=previous_daily_tracker)



from __future__ import annotations

from typing import TYPE_CHECKING

from hyperfocus import formatter
from hyperfocus.commands import SessionHyperfocusCommand, printer
from hyperfocus.database.models import TaskStatus
from hyperfocus.services import DailyTrackerService, PastTrackerService


if TYPE_CHECKING:
    from hyperfocus.session import Session


class NewDayCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)

    def execute(self):
        daily_tracker = DailyTrackerService.from_date(self._session.date)
        if not daily_tracker.is_a_new_day:
            return

        printer.echo(f"✨ {formatter.date(date=self._session.date)}")
        printer.echo("✨ A new day starts, good luck!\n")

        self._session.register_callback(self.review_unfinished_tasks)

    def review_unfinished_tasks(self) -> None:
        daily_tracker = DailyTrackerService.from_date(self._session.date)
        past_tracker = PastTrackerService(daily_tracker)
        prev_day = past_tracker.get_previous_day()
        finished_status = [TaskStatus.DELETED, TaskStatus.DONE]
        if not (tasks := prev_day.get_tasks(exclude=finished_status)):
            return

        printer.echo(f"Unfinished task(s) from {formatter.date(date=prev_day.date)}:")
        printer.tasks(tasks=tasks, newline=True)
        if not printer.confirm(
            f"Review {len(tasks)} unfinished task(s)",
            default=True,
        ):
            return

        for task in tasks:
            if printer.confirm(f'Take back task "{task.title}" for today'):
                daily_tracker.add_task(title=task.title, details=task.details)
        printer.echo("")  # Empty line for design



from abc import ABC

from hyperfocus.session import Session


class HyperfocusCommand(ABC):
    pass


class SessionHyperfocusCommand(HyperfocusCommand, ABC):
    def __init__(self, session: Session) -> None:
        self._session = session



from __future__ import annotations

import datetime
from typing import Callable

import click

from hyperfocus.config.config import Config
from hyperfocus.database import database
from hyperfocus.exceptions import SessionError


class Session:
    _database = database

    def __init__(self) -> None:
        self._config = Config.load()
        self._database.connect(self._config["core.database"])
        self._date = datetime.datetime.now()
        self._callback_commands: list[Callable] = []

    @property
    def config(self) -> Config:
        return self._config

    @property
    def date(self) -> datetime.date:
        return self._date.date()

    @property
    def callback_commands(self) -> list[Callable]:
        return self._callback_commands

    def register_callback(self, callback: Callable) -> None:
        self._callback_commands.append(callback)

    def bind_context(self, ctx: click.Context) -> None:
        ctx.obj = self
        ctx.call_on_close(self.teardown)

    def teardown(self) -> None:
        self._database.close()


def get_current_session() -> Session:
    ctx = click.get_current_context()
    if not isinstance(ctx.obj, Session):
        raise SessionError(
            "It appears that you are trying to invoke a command outside "
            "of the CLI context"
        )

    return ctx.obj



from __future__ import annotations

import functools
from typing import Callable

import click

from hyperfocus.commands import printer
from hyperfocus.exceptions import HyperfocusException, HyperfocusExit
from hyperfocus.utils import un_camel_case


class HyfClickExceptionAdapter(HyperfocusException):
    def __init__(
        self, message: str, event: str, exit_code: int, msg_prefix: str | None = None
    ) -> None:
        super().__init__(message, event)
        self.exit_code = exit_code
        self.msg_prefix = msg_prefix

    @functools.singledispatchmethod
    @classmethod
    def adapt_from(
        cls, exception: click.exceptions.ClickException
    ) -> HyfClickExceptionAdapter:
        return cls(
            message=exception.format_message(),
            event=un_camel_case(type(exception).__name__),
            exit_code=exception.exit_code,
        )

    @adapt_from.register(click.exceptions.UsageError)
    @classmethod
    def _(cls, exception: click.exceptions.UsageError) -> HyfClickExceptionAdapter:
        hint = ""
        usage = ""
        if (
            exception.ctx is not None
            and exception.ctx.command.get_help_option(exception.ctx) is not None
        ):
            command = exception.ctx.command_path
            option = exception.ctx.help_option_names[0]
            hint = f"Try '{command} {option}' for help.\n"
        if exception.ctx is not None:
            usage = f"{exception.ctx.get_usage()}\n"

        return cls(
            message=exception.format_message(),
            event=un_camel_case(type(exception).__name__),
            exit_code=exception.exit_code,
            msg_prefix=f"{usage}{hint}",
        )


def hyf_error_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Callable:
        try:
            return func(*args, **kwargs)
        except HyperfocusException as error:
            printer.error(text=error.message, event=error.event)
            raise HyperfocusExit(error.exit_code)
        except click.ClickException as error:
            hyf_error = HyfClickExceptionAdapter.adapt_from(error)
            if hyf_error.msg_prefix:
                printer.echo(hyf_error.msg_prefix)
            printer.error(text=hyf_error.message, event=hyf_error.event)
            raise HyperfocusExit(error.exit_code)

    return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

import pyperclip

from hyperfocus import formatter
from hyperfocus.commands import SessionHyperfocusCommand, printer
from hyperfocus.exceptions import HyperfocusExit, TaskError
from hyperfocus.services import DailyTrackerService


if TYPE_CHECKING:
    from hyperfocus.database.models import Task, TaskStatus
    from hyperfocus.session import Session


class TaskCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._daily_tracker = DailyTrackerService.from_date(session.date)

    def check_task_id_or_ask(
        self, task_id: int | None, text: str, exclude: list[TaskStatus] | None = None
    ) -> int:
        if task_id:
            return task_id

        exclude = exclude or []
        self.show_tasks(newline=True, exclude=exclude)

        return printer.ask(text, type=int)

    def show_tasks(
        self, exclude: list[TaskStatus] | None = None, newline=False
    ) -> None:
        exclude = exclude or []
        tasks = self._daily_tracker.get_tasks(exclude=exclude)

        if not tasks:
            printer.echo("No tasks for today...")
            raise HyperfocusExit()

        printer.tasks(tasks=tasks, newline=newline)

    def get_task(self, task_id: int) -> Task:
        task = self._daily_tracker.get_task(task_id=task_id)
        if not task:
            raise TaskError(f"Task {task_id} does not exist.")

        return task

    def add_task(self, title: str, details: str) -> Task:
        return self._daily_tracker.add_task(title=title, details=details)

    def update_task(self, task: Task, status: TaskStatus) -> None:
        self._daily_tracker.update_task(task=task, status=status)


class AddTaskCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._task_command = TaskCommand(session)

    def execute(self, title: str, add_details: bool) -> None:
        details = printer.ask("Task details") if add_details else ""

        task = self._task_command.add_task(title=title, details=details)
        printer.success(
            text=formatter.task(task=task, show_prefix=True),
            event="created",
        )


class UpdateTaskCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._task_command = TaskCommand(session)

    def execute(self, task_id: int | None, status: TaskStatus, text: str) -> None:
        task_id = self._task_command.check_task_id_or_ask(
            task_id=task_id, text=text, exclude=[status]
        )

        task = self._task_command.get_task(task_id=task_id)
        if task.status == status.value:
            printer.warning(
                text=formatter.task(task=task, show_prefix=True),
                event="no change",
            )
            raise HyperfocusExit()

        self._task_command.update_task(task=task, status=status)
        printer.success(
            text=formatter.task(task=task, show_prefix=True),
            event="updated",
        )


class ShowTaskCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._task_command = TaskCommand(session)

    def execute(self, task_id: int | None) -> None:
        task_id = self._task_command.check_task_id_or_ask(
            task_id=task_id, text="Show task details"
        )

        task = self._task_command.get_task(task_id=task_id)
        printer.task(task=task, show_details=True, show_prefix=True)


class CopyCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._task_command = TaskCommand(session)

    def execute(self, task_id: int | None) -> None:
        task_id = self._task_command.check_task_id_or_ask(
            task_id=task_id, text="Copy task details"
        )

        task = self._task_command.get_task(task_id=task_id)
        if not task.details:
            raise TaskError(f"Task {task_id} does not have details.")

        pyperclip.copy(task.details)
        printer.success(f"Task {task_id} details copied to clipboard.", event="success")



from __future__ import annotations

from datetime import datetime
from enum import IntEnum, auto

import peewee

from hyperfocus.database import database
from hyperfocus.database.error_handler import db_error_handler
from hyperfocus.utils import wrap_methods


class TaskStatus(IntEnum):
    TODO = auto()
    BLOCKED = auto()
    DELETED = auto()
    DONE = auto()


class BaseModel(peewee.Model):
    class Meta:
        database = database()


@wrap_methods(db_error_handler, ["save", "get_or_create"])
class DailyTracker(BaseModel):
    date = peewee.DateField(primary_key=True, default=datetime.now().date())
    task_increment = peewee.IntegerField(default=0)

    @property
    def next_task_id(self):
        return self.task_increment + 1


@wrap_methods(db_error_handler, ["save", "get_or_none"])
class Task(BaseModel):
    _id = peewee.AutoField()
    id = peewee.IntegerField()
    title = peewee.TextField()
    details = peewee.TextField(null=True)
    status = peewee.IntegerField(default=TaskStatus.TODO)
    postponed_at = peewee.DateField(null=True)
    updated_at = peewee.DateTimeField(default=datetime.now())
    daily_tracker = peewee.ForeignKeyField(DailyTracker, backref="tasks", null=True)

    class Meta:
        table_name = "tasks"
        indexes = ((("id", "daily_tracker"), True),)


MODELS = [DailyTracker, Task]



from __future__ import annotations

import functools

import peewee

from hyperfocus.database.exceptions import DatabaseError


def db_error_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except peewee.DatabaseError as error:
            if "no such table" in str(error):
                raise DatabaseError(
                    "Database not initialized, please run init command first"
                )
            raise DatabaseError(f"Unexpected database error: {error}")

    return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

from hyperfocus.commands import SessionHyperfocusCommand
from hyperfocus.commands.cmd_task import TaskCommand


if TYPE_CHECKING:
    from hyperfocus.session import Session


class StatusCommand(SessionHyperfocusCommand):
    def __init__(self, session: Session) -> None:
        super().__init__(session=session)
        self._task_command = TaskCommand(session)

    def execute(self) -> None:
        self._task_command.show_tasks(newline=True)



from __future__ import annotations

import datetime
from enum import Enum, IntEnum, auto
from typing import List

import click
from tabulate import tabulate

from hyperfocus.database.models import Task, TaskStatus


class NotificationLevel(IntEnum):
    SUCCESS = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()


class Colors(str, Enum):
    BLACK = "black"
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"
    MAGENTA = "magenta"
    CYAN = "cyan"
    WHITE = "white"
    RESET = "reset"
    BRIGHT_BLACK = "bright_black"
    BRIGHT_RED = "bright_red"
    BRIGHT_GREEN = "bright_green"
    BRIGHT_YELLOW = "bright_yellow"
    BRIGHT_BLUE = "bright_blue"
    BRIGHT_MAGENTA = "bright_magenta"
    BRIGHT_CYAN = "bright_cyan"
    BRIGHT_WHITE = "bright_white"


def date(date: datetime.date) -> str:
    return date.strftime("%a, %d %B %Y")


def task_status(status: TaskStatus):
    symbol = "⬢"
    color = {
        TaskStatus.TODO: Colors.WHITE,
        TaskStatus.BLOCKED: Colors.BRIGHT_YELLOW,
        TaskStatus.DELETED: Colors.RED,
        TaskStatus.DONE: Colors.GREEN,
    }.get(status, Colors.BLACK)

    return click.style(symbol, fg=color)


def prompt(text: str):
    symbol = click.style("?", fg=Colors.BRIGHT_GREEN)
    return f"{symbol} {text}"


def task(task: Task, show_details: bool = False, show_prefix: bool = False) -> str:
    empty_details = "No details provided ..."

    title_style = {
        TaskStatus.DELETED: {"fg": Colors.BRIGHT_BLACK},
        TaskStatus.DONE: {"strikethrough": True},
    }.get(task.status, {})

    title = click.style(task.title, **title_style)  # type: ignore
    details = "⊕" if task.details else "◌"
    prefix = f"Task: #{task.id} " if show_prefix else ""

    headline = f"{prefix}{task_status(task.status)} {title}"

    if show_details:
        return f"{headline}\n{task.details or empty_details}"

    return f"{headline} {details}"


def tasks(tasks: List[Task], newline: bool = False) -> str:
    headers = ["#", "tasks"]
    suffix = "\n" if newline else ""
    lines = [[t.id, task(t)] for t in tasks]

    return f"{tabulate(lines, headers)} {suffix}"


def notification(text: str, event: str, status: NotificationLevel) -> str:
    symbol, color = {
        NotificationLevel.SUCCESS: ("✔", Colors.BRIGHT_GREEN),
        NotificationLevel.INFO: ("ℹ", Colors.BRIGHT_CYAN),
        NotificationLevel.WARNING: ("▼", Colors.BRIGHT_YELLOW),
        NotificationLevel.ERROR: ("✘", Colors.BRIGHT_RED),
    }.get(status, (">", Colors.BRIGHT_WHITE))
    prefix = click.style(f"{symbol}({event})", fg=color)

    return f"{prefix} {text}"



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Type

from hyperfocus.config.exceptions import ConfigError


class OptionPolicies:
    def __init__(self, policies: dict[str, Type[OptionPolicy]] | None = None):
        self._policies = policies or {}

    def check(self, section: str, key: str, value: str) -> None:
        policy = self._policies.get(section)
        if policy is not None:
            policy(key, value).validate()


class OptionPolicy(ABC):
    section: str | None = None

    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value

    @property
    def option(self) -> str:
        return f"{self.section}.{self.key}"

    @abstractmethod
    def validate(self) -> None:
        raise NotImplementedError


class CorePolicy(OptionPolicy):
    section = "core"

    def validate(self) -> None:
        if self.key == "database":
            db_path = Path(self.value)
            if not db_path.parent.is_dir():
                raise ConfigError(
                    f"Config option '{self.option}' must be a valid path."
                )
        else:
            raise ConfigError(f"Unknown config option '{self.option}'.")


class AliasPolicy(OptionPolicy):
    section = "alias"

    def validate(self) -> None:
        from hyperfocus import cli

        if self.value not in cli.get_commands():
            raise ConfigError("Alias must referred to an existing command name.")



from pathlib import Path

from hyperfocus.commands import HyperfocusCommand, printer
from hyperfocus.config.config import Config
from hyperfocus.database import database
from hyperfocus.database.models import MODELS


class InitCommand(HyperfocusCommand):
    def execute(self, db_path: str):
        config = self.create_config(db_path=db_path)
        self.init_database(config=config)

    @staticmethod
    def create_config(db_path: str) -> Config:
        config = Config()
        config.make_directory()
        config["core.database"] = str(Path(db_path).resolve())
        config.save()
        printer.info(
            text=f"Config file created successfully in {config.config_file.path}",
            event="init",
        )

        return config

    @staticmethod
    def init_database(config: Config) -> None:
        database.connect(config["core.database"])
        database.init_models(MODELS)
        printer.info(
            text=f"Database initialized successfully in {config['core.database']}",
            event="init",
        )



from __future__ import annotations

import re
from typing import Callable, List


def wrap_methods(decorator: Callable, methods: List[str]):
    def wrapper(cls):
        for method in methods:
            setattr(cls, method, decorator(getattr(cls, method)))
        return cls

    return wrapper


def un_camel_case(camel_case_str: str):
    return re.sub(r"([A-Z])", r" \1", camel_case_str).lower().strip()



from __future__ import annotations

from hyperfocus.commands import SessionHyperfocusCommand, printer


class ConfigCommand(SessionHyperfocusCommand):
    def execute(
        self, option: str | None, value: str | None, list_: bool, unset: bool
    ) -> None:
        if list_:
            self._show_config()
        elif unset and option is not None:
            self._delete_option(option=option)
        elif option is not None and value is not None:
            self._edit_option(option=option, value=value)

    def _show_config(self) -> None:
        for option, value in self._session.config.options.items():
            printer.echo(f"{option} = {value}")

    def _save_config(self) -> None:
        self._session.config.save()
        printer.success("Config updated", event="success")

    def _delete_option(self, option: str) -> None:
        del self._session.config[option]
        self._save_config()

    def _edit_option(self, option: str, value) -> None:
        self._session.config[option] = value
        self._save_config()



from __future__ import annotations

import sys

import click

from hyperfocus import __app_name__, __version__, formatter
from hyperfocus.commands.cmd_config import ConfigCommand
from hyperfocus.commands.cmd_init import InitCommand
from hyperfocus.commands.cmd_new_day import NewDayCommand
from hyperfocus.commands.cmd_status import StatusCommand
from hyperfocus.commands.cmd_task import (
    AddTaskCommand,
    CopyCommand,
    ShowTaskCommand,
    UpdateTaskCommand,
)
from hyperfocus.database.models import TaskStatus
from hyperfocus.hyf_click.core import HyfGroup
from hyperfocus.hyf_click.parameters import NotRequired, NotRequiredIf
from hyperfocus.locations import DEFAULT_DB_PATH
from hyperfocus.session import Session, get_current_session


@click.group(cls=HyfGroup, help="Minimalist task manager")
@click.version_option(
    version=__version__, prog_name=__app_name__, help="Show the version"
)
@click.pass_context
def hyf(ctx: click.Context) -> None:
    if ctx.invoked_subcommand in ["init"] or "--help" in sys.argv[1:]:
        return

    session = Session()
    session.bind_context(ctx=ctx)

    NewDayCommand(session).execute()


@hyf.result_callback()
def process_session(session: Session | None, **_):
    if session is None:
        return

    for callback_commands in session.callback_commands:
        callback_commands()


@hyf.command(help="Initialize hyperfocus config and database")
@click.option(
    "--db-path",
    default=DEFAULT_DB_PATH,
    prompt=formatter.prompt("Database location"),
    help="Database file location",
)
def init(db_path: str) -> None:
    InitCommand().execute(db_path=db_path)


@hyf.command(help="Show current working day status")
def status() -> Session:
    session = get_current_session()
    StatusCommand(session).execute()

    return session


@hyf.command(help="Add task to current working day")
@click.argument("title", metavar="<title>", type=click.STRING)
@click.option("-d", "--details", "add_details", is_flag=True, help="add task details")
def add(title: str, add_details: bool) -> Session:
    session = get_current_session()
    AddTaskCommand(session).execute(title=title, add_details=add_details)

    return session


@hyf.command(help="Mark task as done")
@click.argument("task_id", metavar="<id>", required=False, type=click.INT)
def done(task_id: int | None) -> Session:
    session = get_current_session()
    UpdateTaskCommand(session).execute(
        task_id=task_id, status=TaskStatus.DONE, text="Validate task"
    )

    return session


@hyf.command(help="Reset task as todo")
@click.argument("task_id", metavar="<id>", required=False, type=int)
def reset(task_id: int | None) -> Session:
    session = get_current_session()
    UpdateTaskCommand(session).execute(
        task_id=task_id, status=TaskStatus.TODO, text="Reset task"
    )

    return session


@hyf.command(help="Mark task as blocked")
@click.argument("task_id", metavar="<id>", required=False, type=click.INT)
def block(task_id: int | None) -> Session:
    session = get_current_session()
    UpdateTaskCommand(session).execute(
        task_id=task_id, status=TaskStatus.BLOCKED, text="Block task"
    )

    return session


@hyf.command(help="Delete given task")
@click.argument("task_id", metavar="<id>", required=False, type=click.INT)
def delete(task_id: int | None) -> Session:
    session = get_current_session()
    UpdateTaskCommand(session).execute(
        task_id=task_id, status=TaskStatus.DELETED, text="Delete task"
    )

    return session


@hyf.command(help="Show task details")
@click.argument("task_id", metavar="<id>", required=False, type=click.INT)
def show(task_id: int | None) -> Session:
    session = get_current_session()
    ShowTaskCommand(session).execute(task_id=task_id)

    return session


@hyf.command(help="Copy task details into clipboard")
@click.argument("task_id", metavar="<id>", required=False, type=click.INT)
def copy(task_id: int | None) -> Session:
    session = get_current_session()
    CopyCommand(session).execute(task_id=task_id)

    return session


@hyf.command(help="Get and set options")
@click.argument(
    "option",
    cls=NotRequiredIf,
    not_required_if=["list"],
    metavar="<option>",
    type=click.STRING,
)
@click.argument(
    "value",
    cls=NotRequiredIf,
    not_required_if=["list", "unset"],
    metavar="<value>",
    type=click.STRING,
)
@click.option(
    "--unset",
    cls=NotRequired,
    not_required=["value", "list"],
    is_flag=True,
    help="Unset an option",
)
@click.option(
    "--list",
    "list_",
    cls=NotRequired,
    not_required=["option", "value", "unset"],
    is_flag=True,
    help="Show the whole config",
)
def config(option: str | None, value: str | None, list_: bool, unset: bool) -> Session:
    session = get_current_session()
    ConfigCommand(session).execute(option=option, value=value, list_=list_, unset=unset)

    return session


def get_commands() -> list[str]:
    return [command for command in hyf.commands.keys()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hyperfocus',
 'hyperfocus.commands',
 'hyperfocus.config',
 'hyperfocus.database',
 'hyperfocus.hyf_click']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'peewee>=3.14.8,<4.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'tabulate>=0.8.9,<0.9.0']

entry_points = \
{'console_scripts': ['hyf = hyperfocus.cli:hyf']}

setup_kwargs = {
    'name': 'hyperfocus',
    'version': '0.2.0',
    'description': 'Minimalist task manager.',
    'long_description': '<p align="center">\n    <a href="#readme">\n        <img alt="HyperFocus logo" src="https://raw.githubusercontent.com/u8slvn/hyperfocus/main/_statics/logo.png">\n    </a>\n</p>\n<p align="center">\n    <a href="https://pypi.org/project/hyperfocus/"><img src="https://img.shields.io/pypi/v/hyperfocus.svg" alt="Pypi Version"></a>\n    <a href="https://pypi.org/project/hyperfocus/"><img src="https://img.shields.io/pypi/pyversions/hyperfocus" alt="Python Version"></a>\n    <a href="https://github.com/u8slvn/hyperfocus/actions/workflows/ci.yml"><img src="https://img.shields.io/github/workflow/status/u8slvn/hyperfocus/CI/main" alt="CI"></a>\n    <a href="https://coveralls.io/github/u8slvn/hyperfocus?branch=main"><img src="https://coveralls.io/repos/github/u8slvn/hyperfocus/badge.svg?branch=main" alt="Coverage Status"></a>\n    <a href="https://app.codacy.com/gh/u8slvn/hyperfocus/dashboard"><img src="https://img.shields.io/codacy/grade/01ddd5668dbf4fc09f20ef215d0eec0b" alt="Code Quality"></a>\n    <a href="https://pypi.org/project/hyperfocus/"><img src="https://img.shields.io/pypi/l/hyperfocus" alt="Project license"></a>\n</p>\n\n---\n\n```sh\n$ pipx install hyperfocus\n```\n',
    'author': 'u8slvn',
    'author_email': 'u8slvn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/u8slvn/hyperfocus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Type

from peewee import Model, SqliteDatabase


class _Database:
    def __init__(self) -> None:
        self._engine = SqliteDatabase(None)

    def __call__(self) -> SqliteDatabase:
        return self._engine

    def connect(self, file: str) -> None:
        self._engine.init(file, pragmas={"foreign_keys": 1})

    def init_models(self, models: list[Type[Model]]) -> None:
        with self._engine:
            self._engine.create_tables(models=models)

    def close(self) -> None:
        self._engine.close()



from __future__ import annotations

import click

from hyperfocus.utils import un_camel_case


class HyperfocusException(Exception):
    exit_code = 1

    def __init__(self, message: str, event: str | None = None):
        super().__init__(message)
        self._event = event
        self.message = message

    @property
    def event(self):
        if self._event:
            return self._event

        return un_camel_case(type(self).__name__)


class HyperfocusExit(click.exceptions.Exit):
    pass


class AliasError(HyperfocusException):
    pass


class TaskError(HyperfocusException):
    pass


class SessionError(HyperfocusException):
    pass


class ServiceError(HyperfocusException):
    pass



from __future__ import annotations

import contextlib
import re
from copy import deepcopy
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Generator

from hyperfocus.config.exceptions import ConfigError, ConfigFileError
from hyperfocus.config.file import ConfigFile
from hyperfocus.config.policy import AliasPolicy, CorePolicy, OptionPolicies
from hyperfocus.locations import CONFIG_DIR, CONFIG_PATH


_loaded_config: Config | None = None


class Config:
    default_config: dict[str, dict] = {
        "core": {
            "database": "",
        },
        "alias": {},
    }
    _option_policies = OptionPolicies(
        {
            "core": CorePolicy,
            "alias": AliasPolicy,
        }
    )

    def __init__(self) -> None:
        self._config = deepcopy(self.default_config)
        self._config_file: ConfigFile = ConfigFile(CONFIG_PATH)

    @staticmethod
    def make_directory() -> None:
        try:
            CONFIG_DIR.mkdir(exist_ok=True)
        except OSError as error:
            raise ConfigError(f"Configuration folder creation failed: {error}")

    def merge(self, config: dict[str, Any]) -> None:
        self._config.update(config)

    @property
    def config(self) -> dict[str, dict]:
        return self._config

    @property
    def config_file(self) -> ConfigFile:
        return self._config_file

    def set_config_file(self, config_file: ConfigFile) -> None:
        self._config_file = config_file

    @classmethod
    def load(cls, config_path: Path | None = None, reload: bool = False) -> Config:
        global _loaded_config

        if _loaded_config is None or reload:
            config_path = config_path or CONFIG_PATH
            config_file = ConfigFile(config_path)
            if not config_file.exists():
                raise ConfigError(
                    "Config does not exist, please run init command first."
                )

            _loaded_config = cls()
            _loaded_config.merge(config_file.read())
            _loaded_config.set_config_file(config_file)

        return _loaded_config

    def save(self):
        try:
            self._config_file.write(self._config)
        except ConfigFileError as error:
            raise ConfigError(
                f"Saving config to {self._config_file.path} failed: {error}."
            )

    @property
    def options(self) -> dict[str, str]:
        options = {}
        for section, settings in self._config.items():
            options.update({f"{section}.{i}": v for i, v in settings.items()})
        return options

    def get_option(self, option: str) -> str:
        with secured_option(option=option) as opt:
            return self._config[opt.section][opt.key]

    def update_option(self, option: str, value: str) -> None:
        with secured_option(option=option) as opt:
            self._option_policies.check(section=opt.section, key=opt.key, value=value)
            self._config[opt.section][opt.key] = value

    def delete_option(self, option: str) -> None:
        with secured_option(option=option) as opt:
            del self._config[opt.section][opt.key]

    def __getitem__(self, option: str) -> str:
        return self.get_option(option=option)

    def __setitem__(self, option: str, value: str) -> None:
        self.update_option(option=option, value=value)

    def __delitem__(self, option: str) -> None:
        self.delete_option(option=option)

    def __contains__(self, option: str) -> bool:
        return option in self.options


@contextlib.contextmanager
def secured_option(option: str) -> Generator:
    match = re.match(r"^(?P<section>[A-Za-z0-9]+).(?P<key>[A-Za-z0-9]+)$", option)
    if match is None:
        raise ConfigError(f"Bad format config option '{option}'.")

    try:
        yield SimpleNamespace(**match.groupdict())
    except KeyError:
        raise ConfigError(f"Variable '{option}' does not exist.")
